@@ -13,10 +13,7 @@
 using namespace antlr4;
 using namespace antlr4::tree;
 
-TerminalNodeImpl::TerminalNodeImpl(Token *symbol_) : symbol(symbol_) {
-}
-
-Token* TerminalNodeImpl::getSymbol() {
+Token* TerminalNodeImpl::getSymbol() const {
   return symbol;
 }
 
@@ -33,7 +30,7 @@ misc::Interval TerminalNodeImpl::getSourceInterval() {
   return misc::Interval(tokenIndex, tokenIndex);
 }
 
-antlrcpp::Any TerminalNodeImpl::accept(ParseTreeVisitor *visitor) {
+std::any TerminalNodeImpl::accept(ParseTreeVisitor *visitor) {
   return visitor->visitTerminal(this);
 }
 
@@ -41,7 +38,7 @@ std::string TerminalNodeImpl::getText() {
   return symbol->getText();
 }
 
-std::string TerminalNodeImpl::toStringTree(Parser * /*parser*/) {
+std::string TerminalNodeImpl::toStringTree(Parser * /*parser*/, bool /*pretty*/) {
   return toString();
 }
 
@@ -52,6 +49,6 @@ std::string TerminalNodeImpl::toString() {
   return symbol->getText();
 }
 
-std::string TerminalNodeImpl::toStringTree() {
+std::string TerminalNodeImpl::toStringTree(bool /*pretty*/) {
   return toString();
 }
